#include "im_server.h"

#include <libim/Constants.h>
#include "DeskbarIcon.h"
#include <TranslationUtils.h>

#include <image.h>
#include <Roster.h>
#include <Path.h>
#include <FindDirectory.h>
#include <Directory.h>
#include <stdio.h>
#include <Node.h>
#include <NodeMonitor.h>
#include <VolumeRoster.h>
#include <Volume.h>
#include <UTF8.h>
#include <libim/Helpers.h>
#include <Deskbar.h>
#include <Roster.h>
#include <be/kernel/fs_attr.h>

using namespace IM;

#define AUTOSTART_APPSIG_SETTING "autostart_appsig"

void
_ERROR( const char * error, BMessage * msg )
{
	LOG("im_server", LOW, error, msg);
}

void
_ERROR( const char * error )
{
	_ERROR(error,NULL);
}

void
_SEND_ERROR( const char * text, BMessage * msg )
{
	if ( msg->ReturnAddress().IsValid() )
	{
		BMessage err(ERROR);
		err.AddString("error",text);
		msg->SendReply(&err);
	} else
	{ // no recipient for message replies, write to stdout
		LOG("im_server", LOW, "ERROR: %s",text);
	}
}

/**
	Default constructor. (Starts People-query and ?) loads add-ons
*/
Server::Server()
:	BApplication(IM_SERVER_SIG)
{
	BDirectory dir;
//	These should use find_directory()
	dir.CreateDirectory("/boot/home/config/settings/im_kit", NULL);
	dir.CreateDirectory("/boot/home/config/settings/im_kit/icons", NULL);
	dir.CreateDirectory("/boot/home/config/settings/im_kit/add-ons", NULL);
	dir.CreateDirectory("/boot/home/config/settings/im_kit/add-ons/protocols", NULL);
	
	LoadAddons();
	
	// add deskbar icon
	BDeskbar db;

	if ( db.RemoveItem( DESKBAR_ICON_NAME ) != B_OK )
		LOG("im_server", DEBUG, "Error removing deskbar icon (this is ok..)");
	
	IM_DeskbarIcon * i = new IM_DeskbarIcon();
	
	if ( db.AddItem( i ) != B_OK )
	{ // couldn't add BView, try entry_ref
		entry_ref ref;
		
		if ( be_roster->FindApp(IM_SERVER_SIG,&ref) == B_OK )
		{
			if ( db.AddItem( &ref ) != B_OK )
				LOG("im_server", LOW, "Error adding icon to deskbar!");
		} else
		{
			LOG("im_server", LOW, "be_roster->FindApp() failed");
		}
	}
	
	delete i;
	
	BPath prefsPath;
	find_directory(B_USER_SETTINGS_DIRECTORY, &prefsPath, true);
	prefsPath.Append("im_kit/icons");
	
	BString iconPath = prefsPath.Path();
	iconPath << "/Online";
	
	fIcons.AddPointer(ONLINE_TEXT "_small", (const void *)GetBitmapFromAttribute(
		iconPath.String(), BEOS_SMALL_ICON));
	fIcons.AddPointer(ONLINE_TEXT "_large", (const void *)GetBitmapFromAttribute(
		iconPath.String(), BEOS_LARGE_ICON));
		
	iconPath = prefsPath.Path();
	iconPath << "/Away";
	fIcons.AddPointer(AWAY_TEXT "_small", (const void *)GetBitmapFromAttribute(
		iconPath.String(), BEOS_SMALL_ICON));
	fIcons.AddPointer(AWAY_TEXT "_large", (const void *)GetBitmapFromAttribute(
		iconPath.String(), BEOS_LARGE_ICON));

	iconPath = prefsPath.Path();
	iconPath << "/Offline";
	fIcons.AddPointer(OFFLINE_TEXT "_small", (const void *)GetBitmapFromAttribute(
		iconPath.String(), BEOS_SMALL_ICON));
	fIcons.AddPointer(OFFLINE_TEXT "_large", (const void *)GetBitmapFromAttribute(
		iconPath.String(), BEOS_LARGE_ICON));

	SetAllOffline();

	StartAutostartApps();
	
	Run();
}

/**
	Default destructor. Unloads add-ons and frees any aquired resources.
*/
Server::~Server()
{
	StopAutostartApps();

	UnloadAddons();
	
	SetAllOffline();

	BDeskbar db;
	db.RemoveItem( DESKBAR_ICON_NAME );
}

/**
*/
bool
Server::QuitRequested()
{
	return true;
}

// Loads attribute named 'attribute' from the file 'name'defaults to a type of 'BBMP'
// Doesn't work on symlinks

BBitmap *
Server::GetBitmapFromAttribute(const char *name, const char *attribute, 
	type_code type = 'BBMP') {
	BBitmap 	*bitmap = NULL;
	size_t 		len = 0;
	status_t 	error;	

	if ((name == NULL) || (attribute == NULL)) return NULL;

	BNode node(name);
	
	if (node.InitCheck() != B_OK) {
		return NULL;
	};
	
	attr_info info;
		
	if (node.GetAttrInfo(attribute, &info) != B_OK) {
		node.Unset();
		return NULL;
	};
		
	char *data = (char *)calloc(info.size, sizeof(char));
	len = (size_t)info.size;
		
	if (node.ReadAttr(attribute, 'BBMP', 0, data, len) != len) {
		node.Unset();
		free(data);
	
		return NULL;
	};
	
//	Icon is a square, so it's right / bottom co-ords are the root of the bitmap length
//	Offset is 0
	BRect bound = BRect(0, 0, 0, 0);
	bound.right = sqrt(len) - 1;
	bound.bottom = bound.right;
	
	bitmap = new BBitmap(bound, B_COLOR_8_BIT);
	bitmap->SetBits(data, len, 0, B_COLOR_8_BIT);

//	make sure it's ok
	if(bitmap->InitCheck() != B_OK) {
		free(data);
		delete bitmap;
		return NULL;
	};
	
	return bitmap;
}

BBitmap *
Server::GetBitmap(const char *name, type_code type = 'BBMP') {
	BResources *res = AppResources();

	BBitmap 	*bitmap = NULL;
	size_t 		len = 0;
	status_t 	error;	

	// load resource
	const void *data = res->LoadResource(type, name, &len);
	
	BMemoryIO stream(data, len);
	
	// unflatten it
	BMessage archive;
	error = archive.Unflatten(&stream);
	if (error != B_OK)
		return NULL;

	// make a bbitmap from it
	bitmap = new BBitmap(&archive);
	if(!bitmap)
		return NULL;

	// make sure it's ok
	if(bitmap->InitCheck() != B_OK)
	{
		delete bitmap;
		return NULL;
	}
	
	return bitmap;
}

/**
*/
void
Server::MessageReceived( BMessage * msg )
{
	switch ( msg->what )
	{
		// People messages. Both query and node monitor.
		case B_NODE_MONITOR:
		case B_QUERY_UPDATE:
			HandleContactUpdate(msg);
			break;

		// IM-Kit specified messages:
		
		case ADD_AUTOSTART_APPSIG:
			reply_ADD_AUTOSTART_APPSIG(msg);
			break;
		
		case REMOVE_AUTOSTART_APPSIG:
			reply_REMOVE_AUTOSTART_APPSIG(msg);
			break;
		
		case SERVER_BASED_CONTACT_LIST:
			reply_SERVER_BASED_CONTACT_LIST(msg);
			break;
		
		case GET_SETTINGS_TEMPLATE:
			reply_GET_SETTINGS_TEMPLATE(msg);
			break;
		
		case GET_SETTINGS:
			reply_GET_SETTINGS(msg);
			break;
		
		case SET_SETTINGS:
			reply_SET_SETTINGS(msg);
			break;
		
		case ADD_ENDPOINT:
		{
			BMessenger msgr;
			if ( msg->FindMessenger("messenger",&msgr) == B_OK )
			{
				AddEndpoint( msgr );
				
				msg->SendReply( ACTION_PERFORMED );
			}
		}	break;
		
		case REMOVE_ENDPOINT:
		{
			BMessenger msgr;
			if ( msg->FindMessenger("messenger",&msgr) == B_OK )
			{
				RemoveEndpoint( msgr );
				
				msg->SendReply( ACTION_PERFORMED );
			}
		}	break;
		
		case GET_LOADED_PROTOCOLS:
			reply_GET_LOADED_PROTOCOLS(msg);
			break;
		
		case MESSAGE:
			Process(msg);
			break;
		
		// Other messages
		default:
			BApplication::MessageReceived(msg);
			break;
	}
}

/**
	Initializes People-query and fetches initial list of matches
	
	NOT IMPLEMENTED
*/
void
Server::StartQuery()
{
/*	// make sure we get the whole list before any query updates
	// show up in MessageReceived()
	Lock();
	
	entry_ref entry;
	
	fQuery.SetTarget( BMessenger(this) );
	fQuery.SetPredicate("IM:connections=*");
	
	// get initial list of matches
	while ( fQuery.GetNextRef( &entry ) == B_OK )
	{
		Contact contact(entry);
		
		for ( int i=0; i<contact.CountConnections(); i++ )
		{
			char connection[256];
			if ( contact.ConnectionAt(i,connection) != B_OK )
				continue;
			
			fContactNodes[connection] = contact;
		}
	}
	
	Unlock();
*/
}

/**
*/
void
Server::LoadAddons()
{
	UnloadAddons(); // make sure we don't load any addons twice

	// try loading all files in ./add-ons
	
	// get path
/*	app_info info;
    be_app->GetAppInfo(&info); 

	BPath path;
	BEntry entry(&info.ref); 
	entry.GetPath(&path); 
	path.GetParent(&path);
	path.Append("add-ons");
*/
	BPath path("/boot/home/config/settings/im_kit/add-ons/protocols");
	BEntry entry;
	
	// setup Directory to get list of files
	BDirectory dir( path.Path() );
	
	LOG("im_server", MEDIUM, "add-on directory: %s", path.Path());
	
	while( dir.GetNextEntry( (BEntry*)&entry, TRUE ) == B_NO_ERROR )
	{ // continue until no more files
		if( entry.InitCheck() != B_NO_ERROR )
			continue;
		
		if( entry.GetPath(&path) != B_NO_ERROR )
			continue;
		
		image_id curr_image = load_add_on( path.Path() );
		if( curr_image < 0 )
		{
			LOG("im_server", LOW, "load_add_on() fail, file [%s]", path.Path());
			continue;
		}
		
		status_t res;
		// get name
		Protocol * (* load_protocol) ();
		res = get_image_symbol( 
			curr_image, "load_protocol", 
			B_SYMBOL_TYPE_TEXT, 
			(void **)&load_protocol
		);
		
		if ( res != B_OK )
		{
			LOG("im_server", MEDIUM, "get_image_symbol(load_protocol) fail, file [%s]", path.Path());
			unload_add_on( curr_image );
			continue;
		}
		
		Protocol * protocol = load_protocol();
		
		if ( !protocol )
		{
			LOG("im_server", LOW, "load_protocol() fail, file [%s]", path.Path());
			unload_add_on( curr_image );
		}
		
		LOG("im_server", MEDIUM, "Protocol loaded: [%s]", protocol->GetSignature());
		
		// add to list
		fProtocols[protocol->GetSignature()] = protocol;
		
		// add to fAddOnInfo
		AddOnInfo pinfo;
		pinfo.protocol = protocol;
		pinfo.image = curr_image;
		pinfo.signature = protocol->GetSignature();
		strcpy(pinfo.path,path.Path());
		fAddOnInfo[protocol] = pinfo;
		
		// try to read settings from protocol attribute
		char settings_path[512];
		sprintf(settings_path,"/boot/home/config/settings/im_kit/add-ons/protocols/%s",pinfo.signature);
		BDirectory sdir;
		sdir.CreateFile(settings_path,NULL,true);
		BNode node(settings_path);
		char settings[1024*1024];
		int32 num_read=0;
		
		num_read = node.ReadAttr(
			"im_settings", B_RAW_TYPE, 0,
			settings, sizeof(settings)
		);
		if ( num_read > 0 )
		{
			LOG("im_server", HIGH, "Read settings data");
			
			BMessage settings_msg;
			if ( settings_msg.Unflatten(settings) == B_OK )
			{
				protocol->UpdateSettings(settings_msg);
			} else
			{
				_ERROR("Error unflattening settings");
/*				for ( int i=0; i<num_read; i++ )
				{
					printf("%c",settings[i], settings[i]);
					if ( i%16 == 15 )
						printf("\n");
				}
				printf("\n");
				for ( int i=0; i<num_read; i++ )
				{
					printf("%01x,",settings[i], settings[i]);
					if ( i%16 == 15 )
						printf("\n");
				}
				printf("\n");
*/			}
		} else
		{
			_ERROR("No settings found");
		}
		
		// we should care about the result here..
		protocol->Init( BMessenger(this) );
	}
	LOG("im_server", MEDIUM, "All add-ons loaded.");
}

/**
	Currently does nothing. Will eventually unload add-on images.
*/
void
Server::UnloadAddons()
{
	map<Protocol*,AddOnInfo>::iterator i;
	
	for ( i=fAddOnInfo.begin(); i != fAddOnInfo.end(); i++ )
	{
		i->first->Shutdown();
		
		unload_add_on(i->second.image);
	}
	
	fAddOnInfo.clear();
	fProtocols.clear();
}

/**
*/
void
Server::AddEndpoint( BMessenger msgr )
{
//	printf("Endpoint added\n");
	fMessengers.push_back(msgr);
}

/**
*/
void
Server::RemoveEndpoint( BMessenger msgr )
{
//	printf("Endpoint removed\n");
	fMessengers.remove(msgr);
}

/**
	Process an IM_MESSAGE BMessage, broadcasting or posting to protocols
	as needed to perform the requested operation. Called from MessageReceived()
	
	@param msg The message to process
*/
void
Server::Process( BMessage * msg )
{
	int32 im_what = -1;
	
	if ( msg->FindInt32("im_what",&im_what) != B_OK )
	{ // malformed message, skip
		_ERROR("Server::Process(): Malformed message", msg);
		return;
	}
	
	switch ( im_what )
	{
		// messages to protocols
		case GET_CONTACT_LIST:
		case SET_STATUS:
		case SEND_MESSAGE:
		case GET_CONTACT_INFO:
		{
			MessageToProtocols(msg);
		}	break;
		
		// messages from protocols
		case STATUS_SET:
		case MESSAGE_SENT:
		case MESSAGE_RECEIVED:
		case STATUS_CHANGED:
		case CONTACT_LIST:
		case CONTACT_INFO:
		{
			MessageFromProtocols(msg);
		}	break;
		default:
			// unknown im_what opcode, skip and report
			_ERROR("Unknown im_what code",msg);
			return;
	}
}

/**
	Send a BMessage to all registered BMessengers after converting 'message' encoding if needed.
	
	@param msg The message to send
*/
void
Server::Broadcast( BMessage * msg )
{
	list<BMessenger>::iterator i;
	
	for ( i=fMessengers.begin(); i != fMessengers.end(); i++ )
	{
		if ( !(*i).IsTargetLocal() )
		{
			(*i).SendMessage(msg);
		} else
		{
			_ERROR("Broadcast(): messenger local");
		}
	}
}

/**
	Handle a query update, adding or removing Contacts as needed.
	
	@param msg The message containing the update information
*/
void
Server::HandleContactUpdate( BMessage * msg )
{
	int32 opcode=0;
	
	if ( msg->FindInt32("opcode",&opcode) != B_OK )
	{ // malformed message, skip
		return;
	}
	
	switch ( opcode )
	{
		case B_ENTRY_CREATED:
			// .. add to contact list
			break;
		case B_ENTRY_REMOVED:
			// .. remove from contact list
			break;
	}
}

/**
	Find a Contact associated with a given protocol:id pair
	
	@param proto_id The protocol:id pair to search for
*/
Contact
Server::FindContact( const char * proto_id )
{
	BVolumeRoster 	vroster;
	BVolume			boot_volume;

	vroster.GetBootVolume(&boot_volume);
	
	Contact result;
	
	BQuery query;
	
	string pred = 
		string( 
			string("IM:connections=\"*") + 
			string(proto_id) +
			string("*\"")
		);
	
	query.SetPredicate( pred.c_str() );
	
	
	query.SetVolume(&boot_volume);
	
	query.Fetch();
	
	entry_ref entry;
	
	if ( query.GetNextRef(&entry) == B_OK )
	{
		result.SetTo(entry);
	}
	
	return result;
}

/**
	Create a new People file with a unique name on the form "Unknown contact X"
	and add the specified proto_id to it
	
	@param proto_id The protocol:id connection of the new contact
*/
Contact
Server::CreateContact( const char * proto_id )
{
	LOG("im_server", LOW, "Creating new contact for connection [%s]", proto_id);
	
	Contact result;
	
	BDirectory dir("/boot/home/people");
	BFile file;
	BEntry entry;
	char filename[512];
	
	// make sure that the target directory exists before we try to create
	// new files

	// TO BE ADDED
	
	// create a new 
	for (int i=1; file.InitCheck() != B_OK; i++ )
	{
		sprintf(filename,"Unknown contact %d",i);
		
		dir.CreateFile(filename,&file,true);
	}
	
	if ( dir.FindEntry(filename,&entry) != B_OK )
	{
		LOG("im_server", LOW, "Error: While creating a new contact, dir.FindEntry() failed. filename was [%s]",filename);
		return result;
	}
	
	LOG("im_server", HIGH, "  created file [%s]", filename);
	
	// file created. set type and add connection
	if ( file.WriteAttr(
		"BEOS:TYPE", B_STRING_TYPE, 0,
		"application/x-person", 21
	) != 21 ) 
	{ // error writing type
		entry.Remove();
		_ERROR("Error writing type to created contact");
		return result;
	}
	
	LOG("im_server", DEBUG, "  wrote type");
	
	// file created. set type and add connection
	result.SetTo( entry );
	
	if ( result.AddConnection(proto_id) != B_OK )
	{
		return Contact();
	}
	
	LOG("im_server", DEBUG, "  wrote connection");
	
	if ( result.SetStatus("offline") != B_OK )
	{
		return Contact();
	}
	
	LOG("im_server", DEBUG, "  wrote status");
	
	// post request info about this contact
	BMessage msg(MESSAGE);
	msg.AddInt32("im_what", GET_CONTACT_INFO);
	msg.AddRef("contact", result);
	
	PostMessage(&msg);
	
	LOG("im_server", DEBUG, "  done.");
	
	return result;
}

void
Server::reply_GET_SETTINGS_TEMPLATE( BMessage * msg )
{
	const char * p = msg->FindString("protocol");
			
	if ( !p )
	{
		p = "";
	}
	
	BMessage t;
	
	if ( strlen(p) > 0 )
	{ // protocol settings
		if (fProtocols.find(p) == fProtocols.end() )
		{
			_SEND_ERROR("ERROR: GET_SETTINGS_TEMPLATE: Protocol not loaded",msg);
			return;
		}
		
		Protocol * protocol = fProtocols[p];
			
		t = protocol->GetSettingsTemplate();
	} else
	{ // im_server settings
		t = GenerateSettingsTemplate();
	}
	
	if ( !msg->ReturnAddress().IsValid() )
	{
		_ERROR("Invalid return address in GetSettingsTemplate()", msg);
	}

	msg->SendReply( &t );
}

void
Server::reply_GET_LOADED_PROTOCOLS( BMessage * msg )
{
	BMessage reply( ACTION_PERFORMED );
	
	map<string,Protocol*>::iterator i;
	
	for ( i = fProtocols.begin(); i != fProtocols.end(); i++ )
	{
		reply.AddString("protocol", i->second->GetSignature() );
	}
	
	msg->SendReply( &reply );
}

void
Server::reply_SERVER_BASED_CONTACT_LIST( BMessage * msg )
{
	const char * protocol = msg->FindString("protocol");
	
	if ( !protocol )
	{
		_ERROR("ERROR: Malformed SERVER_BASED_CONTACT_LIST message",msg);
		return;
	}
	
	for ( int i=0; msg->FindString("id",i); i++ )
	{ // for each ID
		const char * id = msg->FindString("id",i);
		
		string proto_id( string(protocol) + string(":") + string(id) );
		
		Contact c = FindContact( proto_id.c_str() );
		
		if ( c.InitCheck() != B_OK )
		{
			CreateContact( proto_id.c_str() );
		}
	}
}

status_t
Server::GetSettings( const char * protocol_sig, BMessage * settings )
{
	char settings_path[512];
	
	// get path to settings file
	if ( !protocol_sig )
	{ // im_server settings
		strcpy(settings_path,"/boot/home/config/settings/im_kit/im_server.settings");
	} else
	{ // protocol settings
		if ( fProtocols.find(protocol_sig) == fProtocols.end() )
		{
			_ERROR("ERROR: GET_SETTINGS: Protocol not loaded");
			return B_ERROR;
		}
		
		Protocol * protocol = fProtocols[protocol_sig];
		
		sprintf(
			settings_path,
			"/boot/home/config/settings/im_kit/add-ons/protocols/%s",
			fAddOnInfo[protocol].signature
		);
	}
	
	char data[1024*1024];
	
	BNode node( settings_path );
	
	int32 num_read = node.ReadAttr(
		"im_settings", B_RAW_TYPE, 0,
		data, sizeof(data)
	);
	
	if ( num_read <= 0 )
	{
		_ERROR("ERROR: GET_SETTINGS: Error reading settings");
		return B_ERROR;
	}
	
	if ( settings->Unflatten(data) != B_OK )
	{
		_ERROR("ERROR: GET_SETTINGS: Error unflattening settings");
		return B_ERROR;
	}
	
	LOG("im_server", DEBUG, "Read settings from file: %s", settings_path);
	
	return B_OK;
}

status_t
Server::SetSettings( const char * protocol_sig, BMessage * settings )
{
	status_t res = B_ERROR;
	char settings_path[512];
		
	if ( settings->what != SETTINGS )
	{
		_ERROR("ERROR: SET_SETTINGS: Malformed settings message");
		return B_ERROR;
	}
			
	// check settings and get path to settings file
	if ( !protocol_sig )
	{ // im_server settings
		res = UpdateOwnSettings(*settings);
		
		strcpy(settings_path,"/boot/home/config/settings/im_kit/im_server.settings");
	} else
	{ // protocol settings
		if ( fProtocols.find(protocol_sig) == fProtocols.end() )
		{
			_ERROR("ERROR: GET_SETTINGS: Protocol not loaded");
			return B_ERROR;
		}
			
		Protocol * protocol = fProtocols[protocol_sig];
			
		res = protocol->UpdateSettings(*settings);
			
		sprintf(
			settings_path,
			"/boot/home/config/settings/im_kit/add-ons/protocols/%s",
			fAddOnInfo[protocol].signature
		);
	}
	
	// check if settings are ok
	if ( res != B_OK )
	{
		_ERROR("ERROR: SET_SETTINGS: Protocol says settings not valid");
		return B_ERROR;
	}
	
	// save settings
	char data[1024*1024];
	int32 data_size=settings->FlattenedSize();
	
	if ( settings->Flatten(data,data_size) != B_OK )
	{ // error flattening message
		_ERROR("ERROR: SET_SETTINGS: Error flattening settings message");
		return B_ERROR;
	}
	
	BDirectory dir;
	dir.CreateFile(settings_path,NULL,true);

	BNode node( settings_path );
			
	if ( node.InitCheck() != B_OK )
	{
		_ERROR("ERROR: SET_SETTINGS: Error opening save file");
		return B_ERROR;
	}
			
	int32 num_written = node.WriteAttr(
		"im_settings", B_RAW_TYPE, 0,
		data, data_size
	);
			
	if ( num_written != data_size )
	{ // error saving settings
		_ERROR("ERROR: SET_SETTINGS: Error saving settings");
		return B_ERROR;
	}

	LOG("im_server", DEBUG, "Wrote settings to file: %s", settings_path);
	
	return B_OK;			
}

void
Server::MessageToProtocols( BMessage * msg )
{
	entry_ref entry;
	
	if ( msg->FindRef("contact",&entry) == B_OK )
	{ // contact present, select protocol and ID
		Contact contact(entry);
		
		if ( contact.InitCheck() != B_OK )
		{ // invalid target
			_SEND_ERROR("Invalid target, no such contact", msg);
			return;
		}
		
		if ( msg->FindString("protocol") == NULL )
		{ // no protocol specified, figure one out
			char connection[255];
			
			if ( contact.ConnectionAt(0,connection) != B_OK )
			{ // trying to send to a Contact with no connections
				_SEND_ERROR("Target contact has no connections", msg);
				return;
			}
			
			// truncate id part of protocol:id pair to get protocol
			for ( int i=0; connection[i]; i++ )
				if ( connection[i] == ':' )
				{
					connection[i] = 0;
					break;
				}
			
			msg->AddString("protocol",connection);
		} // done chosing protocol
		
		if ( msg->FindString("id") == NULL )
		{ // add protocol-specific ID from Contact if not present
			char connection[255];
			
			if ( contact.FindConnection(msg->FindString("protocol"),connection) != B_OK )
			{
				_ERROR("Couldn't get connection for protocol",msg);
				return;
			}
			
			const char * id=NULL;
			
			for ( int i=0; connection[i]; i++ )
				if ( connection[i] == ':' )
				{
					id = &connection[i+1];
					break;
				}
			
			if ( !id )
			{
				_ERROR("Couldn't get ID from connection",msg);
				return;
			}
			
			msg->AddString("id", id );
		}
	} // done selecting protocol and ID
	
	// copy message so we can broadcast it later, with data intact
	BMessage client_side_msg(*msg);
	
	if ( msg->FindString("protocol") == NULL )
	{ // no protocol specified, send to all?
		LOG("im_server", HIGH, "No protocol specified");
		
		int32 im_what=-1;
		msg->FindInt32("im_what", &im_what);
		
		switch ( im_what )
		{ // send these messages to all loaded protocols
			case SET_STATUS:
			{
				LOG("im_server", HIGH, "  SET_STATUS - Sending to all protocols");
				map<string,Protocol*>::iterator i;
				
				for ( i=fProtocols.begin(); i != fProtocols.end(); i++ )
				{
					if ( i->second->Process(msg) != B_OK )
					{
						_ERROR("Protocol reports error processing message");
					}
				}
			}	break;
			default:
				_ERROR("Invalid message", msg);
				return;
		}
	} else
	{ // protocol mapped
		if ( fProtocols.find(msg->FindString("protocol")) == fProtocols.end() )
		{ // invalid protocol, report and skip
			_ERROR("Protocol not loaded or not installed", msg);
			
			printf("looking for [%s], loaded protocols:\n", msg->FindString("protocol") );
			
			map<string,Protocol*>::iterator i;
			
			for ( i=fProtocols.begin(); i != fProtocols.end(); i++ )
			{
				printf("  [%s]\n", i->first.c_str());
			}
			
			printf("  end of list.\n");
			
			_SEND_ERROR("Protocol not loaded or not installed", msg);
			return;
		}
		
		Protocol * p = fProtocols[msg->FindString("protocol")];		
		
		if ( p->GetEncoding() != 0xffff )
		{ // convert to desired charset
			int32 charset = p->GetEncoding();
			
			msg->AddInt32("charset", charset );
			client_side_msg.AddInt32("charset", charset);
			
			type_code _type;
			char * name;
			int32 count;

#if B_BEOS_VERSION > B_BEOS_VERSION_5			
			for ( int i=0; msg->GetInfo(B_STRING_TYPE, i, (const char **)&name, &_type, &count) == B_OK; i++ )
#else
			for ( int i=0; msg->GetInfo(B_STRING_TYPE, i, &name, &_type, &count) == B_OK; i++ )
#endif
			{ // get string names
				for ( int x=0; x<count; x++ )
				{ // replace all matching strings
					const char * data = msg->FindString(name,x);
					
					int32 src_len = strlen(data);
					int32 dst_len = strlen(data)*5;
					int32 state = 0;
					
					char * new_data = new char[dst_len];
					
					if ( convert_from_utf8(
						charset,
						data,		&src_len,
						new_data,	&dst_len,
						&state				
					) == B_OK )
					{
						new_data[dst_len] = 0;
						
						msg->ReplaceString(name,x,new_data);
					}
					
					delete[] new_data;
				}
			}
		} // done converting charset
		
		if ( p->Process(msg) != B_OK )
		{
			_SEND_ERROR("Protocol reports error processing message", msg);
			return;
		}
	}
	
	// broadcast client_side_msg, since clients want utf8 text, and that has
	// been replaced in msg with protocol-specific data
	Broadcast( &client_side_msg );
}

void
Server::MessageFromProtocols( BMessage * msg )
{
	const char * protocol = msg->FindString("protocol");

	// convert strings to utf8
	type_code _type;
	char * name;
	int32 count;
	int32 charset;
	
	if ( msg->FindInt32("charset",&charset) == B_OK )
	{ // charset present, convert all strings
#if B_BEOS_VERSION > B_BEOS_VERSION_5
		for ( int i=0; msg->GetInfo(B_STRING_TYPE, i, (const char **)&name, &_type, &count) == B_OK; i++ )
#else
		for ( int i=0; msg->GetInfo(B_STRING_TYPE, i, &name, &_type, &count) == B_OK; i++ )
#endif
		{ // get string names
			for ( int x=0; x<count; x++ )
			{ // replace all matching strings
				const char * data = msg->FindString(name,x);
				
				int32 src_len = strlen(data);
				int32 dst_len = strlen(data)*5;
				int32 state = 0;
				
				char * new_data = new char[dst_len];
			
				if ( convert_to_utf8(
					charset,
					data,		&src_len,
					new_data,	&dst_len,
					&state				
				) == B_OK )
				{
					new_data[dst_len] = 0;
			
					msg->ReplaceString(name,x,new_data);
				}
				
				delete[] new_data;
			}
		}
	}	
	// done converting
	
	int32 im_what;
	
	msg->FindInt32("im_what",&im_what);
	
	entry_ref testc;
	
	if ( msg->FindRef("contact",&testc) == B_OK )
	{
		_ERROR("contact already present in message supposedly from protocol",msg);
		return;
	}
	
	// find out which contact this message originates from
	Contact contact;
	
	const char * id = msg->FindString("id");
	
	if ( id != NULL )
	{ // ID present, find out which Contact it belongs to
		if ( protocol == NULL )
		{ // malformed message. report and skip.
			_ERROR("Malformed message in Server::Process", msg);
			return;
		}
		
		string proto_id( string(protocol) + string(":") + string(id) );
		
		contact = FindContact(proto_id.c_str());
		
		if ( contact.InitCheck() != B_OK )
		{ // No matching contact, create a new one!
			contact.SetTo( CreateContact( proto_id.c_str() ) );
		}
	}
	
	if ( contact.InitCheck() == B_OK )
	{
		msg->AddRef("contact",contact);
	}
	
	if ( im_what == STATUS_CHANGED && protocol != NULL && id != NULL )
	{ // update status list on STATUS_CHANGED
		UpdateStatus(msg,contact);
	}
	
	if ( im_what == STATUS_SET && protocol != NULL )
	{ // own status set for protocol, register the id's we're interested in
		if ( !msg->FindString("status") )
		{
			_ERROR("ERROR: STATUS_SET: status not in message",msg);
			return;
		}
		
		if ( strcmp(ONLINE_TEXT,msg->FindString("status")) == 0 )
		{ // we're online. register contacts. (should be: only do this if we were offline)
			if ( fProtocols.find(protocol) == fProtocols.end() )
			{
				_ERROR("ERROR: STATUS_SET: Protocol not loaded",msg);
				return;
			}
			
			Protocol * p = fProtocols[protocol];
			
			BMessage connections(MESSAGE);
			connections.AddInt32("im_what",REGISTER_CONTACTS);
			GetContactsForProtocol( p->GetSignature(), &connections );
			
			p->Process( &connections );
		}
	}
	
	// send it
	Broadcast(msg);
}

void
Server::UpdateStatus( BMessage * msg, Contact & contact )
{
	const char * status = msg->FindString("status");
	const char * protocol = msg->FindString("protocol");
	const char * id = msg->FindString("id");
	
	if ( !status )
	{
		_ERROR("Missing 'status' in STATUS_CHANGED message",msg);
		return;
	}
	
	string proto_id( string(protocol) + string(":") + string(id) );
	
	string new_status = status;
	
	LOG("im_server", MEDIUM, "STATUS_CHANGED [%s] is now %s",proto_id.c_str(),new_status.c_str());
	
	// update status
	fStatus[proto_id] = new_status;

	// calculate total status for contact
/*	for ( int i=0; i<contact.CountConnections(); i++ )
	{ // calc total status
		char connection[512];
		
		contact.ConnectionAt(i,connection);
		
		if ( proto_id != connection )
		{ // not this connection, check status
			string curr = fStatus[connection];
			
			if ( curr == AWAY_TEXT && new_status == OFFLINE_TEXT )
			{
				new_status = AWAY_TEXT;
			}
			if ( curr == ONLINE_TEXT )
			{
				new_status = ONLINE_TEXT;
			}
		}
	}
*/	
	// update status attribute
	BNode node(contact);
	
	if ( node.InitCheck() != B_OK )
	{
		_ERROR("ERROR: Invalid node when setting new status", msg);
	} else
	{ // node exists, write status
		status = new_status.c_str();
		
		if ( node.WriteAttr(
			"IM:status", B_STRING_TYPE, 0,
			status, strlen(status)+1
		) != (int32)strlen(status)+1 )
		{
			_ERROR("Error writing status attribute",msg);
		}
			
		BBitmap *large = NULL;
		BBitmap *small = NULL;
		
		BString pointerName = status;
		pointerName << "_small";

		fIcons.FindPointer(pointerName.String(), reinterpret_cast<void **>(&small));
		
		pointerName = status;
		pointerName << "_large";
		
		fIcons.FindPointer(pointerName.String(), reinterpret_cast<void **>(&large));
		
		if (large != NULL) {
			node.WriteAttr(BEOS_LARGE_ICON, 'ICON', 0, large->Bits(), 
				large->BitsLength());
		} else {
			node.RemoveAttr(BEOS_LARGE_ICON);
		};	

		if (small != NULL) {
			node.WriteAttr(BEOS_SMALL_ICON, 'MICN', 0, small->Bits(), 
				small->BitsLength());
		} else {
			node.RemoveAttr(BEOS_SMALL_ICON);
		};
	}
}

/**
	Query for all files with a IM:status and set it to OFFLINE_TEXT
*/
void
Server::SetAllOffline()
{
	BVolumeRoster 	vroster;
	BVolume			boot_volume;

	vroster.GetBootVolume(&boot_volume);
	
	BQuery query;
	
	query.SetPredicate( "IM:connections=*" );
	query.SetVolume(&boot_volume);
	query.Fetch();
	
	entry_ref entry;
	
	BMessage msg;
	
	while ( query.GetNextRef(&entry) == B_OK )
	{
		msg.AddRef("contact",&entry);
	}
	
	char nickname[512], name[512], filename[512];
	
	Contact c;
	for ( int i=0; msg.FindRef("contact",i,&entry) == B_OK; i++ )
	{
		c.SetTo(&entry);
		
		if ( c.InitCheck() != B_OK )
			_ERROR("SetAllOffline: Contact invalid");
		
		if ( c.GetNickname(nickname,sizeof(nickname)) != B_OK )
			strcpy(nickname,"<no nick>");
		if ( c.GetName(name,sizeof(name)) != B_OK )
			strcpy(name,"<no name>");
		BEntry e(&entry);
		if ( e.GetName(filename) != B_OK )
			strcpy(filename,"<no filename?!>");
			
		LOG("im_server", DEBUG, "Setting %s (%s) offline, filename: %s", name, nickname, filename);
		
		if ( c.SetStatus(OFFLINE_TEXT) != B_OK )
			LOG("im_server", DEBUG, "  error.");

		BNode node(&entry);

		BBitmap *large = NULL;
		BBitmap *small = NULL;
		
		fIcons.FindPointer(OFFLINE_TEXT "_small", reinterpret_cast<void **>(&small));
		fIcons.FindPointer(OFFLINE_TEXT "_large", reinterpret_cast<void **>(&large));
		
		if (large != NULL) {
			node.WriteAttr(BEOS_LARGE_ICON, 'ICON', 0, large->Bits(), 
				large->BitsLength());
		} else {
			node.RemoveAttr(BEOS_LARGE_ICON);
		};	

		if (small != NULL) {
			node.WriteAttr(BEOS_SMALL_ICON, 'MICN', 0, small->Bits(), 
				small->BitsLength());
		} else {
			node.RemoveAttr(BEOS_SMALL_ICON);
		};
	}
}

/**
	Add the ID of all contacts that have a connections for this protocol
	to msg.
*/
void
Server::GetContactsForProtocol( const char * protocol, BMessage * msg )
{
	BVolumeRoster 	vroster;
	BVolume			boot_volume;
	
	vroster.GetBootVolume(&boot_volume);
	
	Contact result;
	
	BQuery query;
	
	char pred[256];
	
	sprintf(pred,"IM:connections=*%s*",protocol);
	
	query.SetPredicate( pred );
	query.SetVolume( &boot_volume );
	query.Fetch();
	
	entry_ref entry;
	
	while ( query.GetNextRef(&entry) == B_OK )
	{
		Contact c(&entry);
		char conn[256];
		char * id = NULL;
		
		if ( c.FindConnection(protocol,conn) == B_OK )
		{
			for ( int i=0; conn[i]; i++ )
				if ( conn[i] == ':' )
				{
					id = &conn[i+1];
					break;
				}
			
			msg->AddString("id",id);
		}
	}
}

BMessage
Server::GenerateSettingsTemplate()
{
	BMessage main_msg(SETTINGS_TEMPLATE);
	
	BMessage blink_db;
	blink_db.AddString("name", "blink_db");
	blink_db.AddString("description", "Blink Deskbar icon");
	blink_db.AddInt32("type", B_BOOL_TYPE );
	blink_db.AddBool("default", true );
	
	main_msg.AddMessage("setting", &blink_db);
	
	return main_msg;
}

status_t
Server::UpdateOwnSettings( BMessage settings )
{
	return B_OK;
}

void
Server::StartAutostartApps()
{
	BMessage settings;
	
	GetSettings( NULL, &settings );
	
	for ( int i=0; settings.FindString(AUTOSTART_APPSIG_SETTING,i); i++ )
	{
		LOG("im_server", LOW, "Starting app [%s]", settings.FindString(AUTOSTART_APPSIG_SETTING,i) );
		be_roster->Launch( settings.FindString(AUTOSTART_APPSIG_SETTING,i) );
	}
}

void
Server::StopAutostartApps()
{
	BMessage settings;
	
	GetSettings( NULL, &settings );
	
	for ( int i=0; settings.FindString(AUTOSTART_APPSIG_SETTING,i); i++ )
	{
		LOG("im_server", LOW, "Stopping app [%s]", settings.FindString(AUTOSTART_APPSIG_SETTING,i) );
		BMessenger msgr( settings.FindString(AUTOSTART_APPSIG_SETTING,i) );
		msgr.SendMessage( B_QUIT_REQUESTED );
	}
}

void
Server::reply_ADD_AUTOSTART_APPSIG( BMessage * msg )
{
	if ( msg->FindString("app_sig") == NULL )
	{
		_SEND_ERROR("No app_sig provided", msg);
		return;
	}
	
	const char * new_appsig = msg->FindString("app_sig");
	
	BMessage settings;
	if ( GetSettings(NULL, &settings) != B_OK )
	{
		_SEND_ERROR("Error reading settings",msg);
		return;
	}
	
	for ( int i=0; settings.FindString(AUTOSTART_APPSIG_SETTING, i); i++ )
	{
		if ( strcmp( new_appsig, settings.FindString(AUTOSTART_APPSIG_SETTING,i) ) == 0 )
		{ // app-sig already present, don't add again
			msg->SendReply(ACTION_PERFORMED);
			LOG("im_server", HIGH, "Auto-start app already present [%s]", new_appsig);
			return;
		}
	}
	
	settings.AddString(AUTOSTART_APPSIG_SETTING, new_appsig);
	
	SetSettings( NULL, &settings );
	
	msg->SendReply(ACTION_PERFORMED);
	
	LOG("im_server", HIGH, "Auto-start app added [%s]", new_appsig);
}

void
Server::reply_REMOVE_AUTOSTART_APPSIG( BMessage * msg )
{
	if ( msg->FindString("app_sig") == NULL )
		_SEND_ERROR("No app_sig provided", msg);
		return;
			
	const char * appsig = msg->FindString("app_sig");
			
	BMessage settings;
	BMessage temp;
	GetSettings(NULL, &settings);
	
	// save app-sigs not to be deleted in temp
	for ( int i=0; settings.FindString(AUTOSTART_APPSIG_SETTING, i); i++ )
	{
		if ( strcmp( appsig, settings.FindString(AUTOSTART_APPSIG_SETTING,i) ) != 0 )
		{ // this is not the app-sig to delete, store it
			temp.AddString(AUTOSTART_APPSIG_SETTING, settings.FindString(AUTOSTART_APPSIG_SETTING,i));
		}
	}
	
	// delete all app-sigs from settings
	settings.RemoveName(AUTOSTART_APPSIG_SETTING);
	
	// copy app-sigs from temp
	for ( int i=0; temp.FindString(AUTOSTART_APPSIG_SETTING,i); i++ )
		settings.AddString(AUTOSTART_APPSIG_SETTING, temp.FindString(AUTOSTART_APPSIG_SETTING,i));
	
	SetSettings( NULL, &settings );

	msg->SendReply(ACTION_PERFORMED);
	
	LOG("im_server", HIGH, "Auto-start app removed [%s]", appsig);
}

void
Server::reply_GET_SETTINGS( BMessage * msg )
{
	const char * protocol = msg->FindString("protocol");
	
	if ( protocol[0] == 0 )
		protocol = NULL;
	
	BMessage settings;
	
	if ( GetSettings( protocol, &settings ) != B_OK )
	{
		_SEND_ERROR("Error getting settings", msg);
	}
	
	msg->SendReply(&settings);
}

void
Server::reply_SET_SETTINGS( BMessage * msg )
{
	const char * protocol = msg->FindString("protocol");
	
	if ( protocol[0] == 0 )
		protocol = NULL;
	
	BMessage settings;
	
	if ( msg->FindMessage("settings", &settings) != B_OK )
	{
		_SEND_ERROR("No settings provided", msg);
		return;
	}
	
	if ( SetSettings( protocol, &settings ) != B_OK )
	{
		_SEND_ERROR("Error setting settings", msg);
	}
	
	msg->SendReply(ACTION_PERFORMED);
}
